mod assets;
mod ui;
use std::{iter::once, sync::Once};

use assets::AssetsController;
use axum::{http, response::IntoResponse, routing::get};
use http::Response;
use hypertext::{maud, GlobalAttributes, Renderable};
use tower_http::{
    compression::CompressionLayer,
    propagate_header::PropagateHeaderLayer,
    sensitive_headers::SetSensitiveRequestHeadersLayer,
    set_header::SetResponseHeaderLayer,
    trace::TraceLayer,
};
use ui::UiController;

use crate::ui::html_elements;

fn markup_404(uri: String) -> impl Renderable {
    maud! {
        h1 { "404" }
        p { (uri) " Not Found" }
        @for i in 0..5 {
            div .{"m-" (i)} { (i) }
        }
    }
}

fn markup_405() -> impl Renderable {
    maud! {
        h1 { "405" }
        p { "Method not allowed!" }
        @for i in 1..3 {
            div .{"m-" (i)} {}
        }
    }
}

async fn handle_404(uri: http::Uri) -> impl IntoResponse {
    (
        http::StatusCode::NOT_FOUND,
        axum::response::Html(markup_404(format!("{uri:?}")).render()),
    )
        .into_response()
}

async fn handle_405() -> impl IntoResponse {
    (
        http::StatusCode::METHOD_NOT_ALLOWED,
        axum::response::Html(markup_405().render()),
    )
        .into_response()
}

/// Initialize the application logger. This should be called once at the start of the application.
/// This function is idempotent and will only initialize the logger once.
pub fn initialize_logger() {
    static INIT: Once = Once::new();

    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::DEBUG.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .pretty()
            .with_env_filter(env_filter)
            .init();
    });
}

#[derive(Clone, Debug)]
struct AppState {
    _field: String,
}

#[tokio::main]
async fn main() {
    initialize_logger();

    let app_state = AppState {
        _field: "".to_string(),
    };

    // TODO pick free port/config
    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();

    fn content_length_from_response<B>(response: &Response<B>) -> Option<http::HeaderValue>
    where
        B: axum::body::HttpBody,
    {
        response
            .body()
            .size_hint()
            .exact()
            .map(|size| http::HeaderValue::from_str(&size.to_string()).unwrap())
    }

    tracing::info!("Listening on http://0.0.0.0:8000");

    let router: axum::Router = axum::Router::new()
        .merge(UiController::into_router(app_state.clone()))
        .merge(AssetsController::into_router(app_state.clone()))
        .fallback_service(get(handle_404))
        .method_not_allowed_fallback(handle_405)
        // High level logging of requests and responses
        .layer(TraceLayer::new_for_http())
        // Mark the `Authorization` request header as sensitive so it doesn't show in logs
        .layer(SetSensitiveRequestHeadersLayer::new(once(http::header::AUTHORIZATION)))
        // Compress responses
        .layer(CompressionLayer::new())
        // Propagate `X-Request-Id`s from requests to responses
        .layer(PropagateHeaderLayer::new(http::HeaderName::from_static("x-request-id")))
        // If the response has a known size set the `Content-Length` header
        .layer(SetResponseHeaderLayer::overriding(http::header::CONTENT_TYPE, content_length_from_response))
        .with_state(app_state);

    let _ = axum::serve(listener, router.into_make_service()).await;
}
